#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import argparse


def resolve_yaml_path(path):
    # priority, filepath > classname > default
    import os
    import io
    if hasattr(path, 'read'):
        # already a readable stream
        return path
    elif os.path.exists(path):
        return open(path, encoding='utf8')
    elif path.isidentifier():
        # possible class name
        return io.StringIO('!%s {}' % path)
    elif path.startswith('!'):
        # possible YAML content
        return io.StringIO(path)
    else:
        raise argparse.ArgumentTypeError('%s can not be resolved, it should be a readable stream,'
                                         ' or a valid file path, or a supported class name.' % path)


def set_base_parser():
    from .. import __version__
    from termcolor import colored
    # create the top-level parser
    parser = argparse.ArgumentParser(
        description='%s, a cloud-native semantic search system '
                    'based on deep neural network. '
                    'It enables large-scale index and semantic search for text-to-text, image-to-image, '
                    'video-to-video and any content form. Visit %s for tutorials and documentations.' % (
                        colored('GNES v%s: Generic Neural Elastic Search' % __version__, 'green'),
                        colored('https://gnes.ai', 'cyan', attrs=['underline'])))
    parser.add_argument('-v', '--version', action='version', version='%(prog)s ' + __version__)
    parser.add_argument('--verbose', action='store_true', default=False,
                        help='turn on detailed logging for debug')
    return parser


def set_composer_parser(parser=None):
    from pkg_resources import resource_stream

    if not parser:
        parser = set_base_parser()
    parser.add_argument('--port',
                        type=int,
                        default=8800,
                        help='host port of the grpc service')
    parser.add_argument('--name',
                        type=str,
                        default='GNES app',
                        help='name of the instance')
    parser.add_argument('--yaml_path', type=resolve_yaml_path,
                        default=resource_stream(
                            'gnes', '/'.join(('resources', 'compose', 'gnes-example.yml'))),
                        help='yaml config of the service')
    parser.add_argument('--html_path', type=argparse.FileType('w', encoding='utf8'),
                        help='output path of the HTML file, will contain all possible generations')
    parser.add_argument('--shell_path', type=argparse.FileType('w', encoding='utf8'),
                        help='output path of the shell-based starting script')
    parser.add_argument('--swarm_path', type=argparse.FileType('w', encoding='utf8'),
                        help='output path of the docker-compose file for Docker Swarm')
    parser.add_argument('--k8s_path', type=argparse.FileType('w', encoding='utf8'),
                        help='output path of the docker-compose file for Docker Swarm')
    parser.add_argument('--graph_path', type=argparse.FileType('w', encoding='utf8'),
                        help='output path of the mermaid graph file')
    parser.add_argument('--shell_log_redirect', type=str,
                        help='the file path for redirecting shell output. '
                             'when not given, the output will be flushed to stdout')
    parser.add_argument('--mermaid_leftright', action='store_true', default=False,
                        help='showing the flow in left-to-right manner rather than top down')
    parser.add_argument('--docker_img', type=str,
                        default='gnes/gnes:latest',
                        help='the docker image used in Docker Swarm & Kubernetes')
    return parser


def set_composer_flask_parser(parser=None):
    if not parser:
        parser = set_base_parser()
    set_composer_parser(parser)
    parser.add_argument('--flask', action='store_true', default=False,
                        help='using Flask to serve a composer in interactive mode, aka GNES board')
    parser.add_argument('--http_port', type=int, default=8080,
                        help='server port for receiving HTTP requests')
    return parser


def set_service_parser(parser=None):
    from ..service.base import SocketType, BaseService, ParallelType
    import random
    if not parser:
        parser = set_base_parser()
    min_port, max_port = 49152, 65536
    parser.add_argument('--port_in', type=int, default=random.randrange(min_port, max_port),
                        help='port for input data, default a random port between [49152, 65536]')
    parser.add_argument('--port_out', type=int, default=random.randrange(min_port, max_port),
                        help='port for output data, default a random port between [49152, 65536]')
    parser.add_argument('--host_in', type=str, default=BaseService.default_host,
                        help='host address for input')
    parser.add_argument('--host_out', type=str, default=BaseService.default_host,
                        help='host address for output')
    parser.add_argument('--socket_in', type=SocketType.from_string, choices=list(SocketType),
                        default=SocketType.PULL_BIND,
                        help='socket type for input port')
    parser.add_argument('--socket_out', type=SocketType.from_string, choices=list(SocketType),
                        default=SocketType.PUSH_BIND,
                        help='socket type for output port')
    parser.add_argument('--port_ctrl', type=int, default=random.randrange(min_port, max_port),
                        help='port for controlling the service, default a random port between [49152, 65536]')
    parser.add_argument('--timeout', type=int, default=-1,
                        help='timeout (ms) of all communication, -1 for waiting forever')
    parser.add_argument('--dump_interval', type=int, default=5,
                        help='serialize the service to a file every n seconds')
    parser.add_argument('--read_only', action='store_true', default=False,
                        help='do not allow the service to modify the model, '
                             'dump_interval will be ignored')
    parser.add_argument('--parallel_backend', type=str, choices=['thread', 'process'], default='thread',
                        help='parallel backend of the service')
    parser.add_argument('--num_parallel', type=int, default=1,
                        help='number of parallel services running at the same time, '
                             '`port_in` and `port_out` will be set to random, '
                             'and routers will be added automatically when necessary')
    parser.add_argument('--parallel_type', type=ParallelType.from_string, choices=list(ParallelType),
                        default=ParallelType.PUSH_NONBLOCK,
                        help='parallel type of the concurrent services')
    return parser


def _set_client_parser(parser=None):
    from ..service.base import SocketType
    if not parser:
        parser = set_base_parser()
    set_service_parser(parser)
    parser.set_defaults(
        port_in=parser.get_default('port_out'),
        port_out=parser.get_default('port_in'),
        socket_in=SocketType.PULL_CONNECT,
        socket_out=SocketType.PUSH_CONNECT,
        read_only=True)
    return parser


def set_loadable_service_parser(parser=None):
    if not parser:
        parser = set_base_parser()
    from ..service.base import SocketType
    set_service_parser(parser)

    parser.add_argument('--yaml_path', type=resolve_yaml_path, required=True,
                        help='yaml config of the service, it should be a readable stream,'
                             ' or a valid file path, or a supported class name.')

    parser.set_defaults(socket_in=SocketType.PULL_BIND,
                        socket_out=SocketType.PUSH_BIND)
    return parser


def set_preprocessor_service_parser(parser=None):
    if not parser:
        parser = set_base_parser()
    set_loadable_service_parser(parser)
    parser.set_defaults(read_only=True)
    return parser


def set_router_service_parser(parser=None):
    if not parser:
        parser = set_base_parser()
    set_loadable_service_parser(parser)
    parser.add_argument('--num_part', type=int, default=None,
                        help='explicitly set the number of parts of message')
    parser.set_defaults(read_only=True)
    return parser


def set_indexer_service_parser(parser=None):
    from ..service.base import SocketType

    if not parser:
        parser = set_base_parser()
    set_loadable_service_parser(parser)

    # encoder's port_out is indexer's port_in
    parser.set_defaults(port_in=parser.get_default('port_out'),
                        port_out=parser.get_default('port_out') + 2,
                        socket_in=SocketType.PULL_CONNECT,
                        socket_out=SocketType.PUB_BIND)
    return parser


def _set_grpc_parser(parser=None):
    if not parser:
        parser = set_base_parser()
    parser.add_argument('--grpc_host',
                        type=str,
                        default='0.0.0.0',
                        help='host address of the grpc service')
    parser.add_argument('--grpc_port',
                        type=int,
                        default=8800,
                        help='host port of the grpc service')
    parser.add_argument('--max_message_size', type=int, default=100,
                        help='maximum send and receive size for grpc server in (MB)')
    return parser


def set_grpc_service_parser(parser=None):
    if not parser:
        parser = set_base_parser()
    set_service_parser(parser)
    _set_grpc_parser(parser)
    parser.add_argument('--pb2_path',
                        type=str,
                        required=True,
                        help='the path of the python file protocol buffer compiler')
    parser.add_argument('--pb2_grpc_path',
                        type=str,
                        required=True,
                        help='the path of the python file generated by the gRPC Python protocol compiler plugin')
    parser.add_argument('--stub_name',
                        type=str,
                        required=True,
                        help='the name of the gRPC Stub')
    parser.add_argument('--api_name',
                        type=str,
                        required=True,
                        help='the api name for calling the stub')
    return parser


def set_frontend_parser(parser=None):
    from ..service.base import SocketType
    if not parser:
        parser = set_base_parser()
    set_service_parser(parser)
    _set_grpc_parser(parser)
    parser.set_defaults(socket_in=SocketType.PULL_BIND,
                        socket_out=SocketType.PUSH_BIND,
                        read_only=True)
    parser.add_argument('--max_concurrency', type=int, default=10,
                        help='maximum concurrent client allowed')
    return parser


def set_cli_client_parser(parser=None):
    import sys
    if not parser:
        parser = set_base_parser()
    _set_grpc_parser(parser)
    parser.add_argument('--txt_file', type=argparse.FileType('r'),
                        default=sys.stdin,
                        help='text file to be used, each line is a doc/query')
    parser.add_argument('--batch_size', type=int, default=100,
                        help='the size of the request to split')
    parser.add_argument('--mode', choices=['index', 'query', 'train'], type=str,
                        required=True,
                        help='the mode of the client and the server')
    parser.add_argument('--data_type', choices=['text', 'image', 'video'], type=str,
                        required=True,
                        help='type of data, available choice: text, image, video')
    parser.add_argument('--image_zip_file', type=str,
                        help='image zip file to be used, consists of multiple images')
    parser.add_argument('--video_zip_file', type=str,
                        help='video zip file to be used, consists of multiple videos')
    parser.add_argument('--top_k', type=int,
                        default=10,
                        help='default top_k for query mode')
    return parser


def set_http_service_parser(parser=None):
    if not parser:
        parser = set_base_parser()
    _set_grpc_parser(parser)
    parser.add_argument('--http_port', type=int, default=80,
                        help='http port to deploy the service')
    parser.add_argument('--http_host', type=str, default='0.0.0.0',
                        help='http host to deploy the service')
    parser.add_argument('--max_workers', type=int, default=100,
                        help='max workers to deal with the message')
    parser.add_argument('--top_k', type=int, default=10,
                        help='default top_k for query mode')
    parser.add_argument('--batch_size', type=int, default=2560,
                        help='batch size for feed data for train mode')
    return parser


def get_main_parser():
    # create the top-level parser
    parser = set_base_parser()
    sp = parser.add_subparsers(dest='cli', title='GNES sub-commands',
                               description='use "gnes [sub-command] --help" '
                                           'to get detailed information about each sub-command')

    # microservices
    set_frontend_parser(sp.add_parser('frontend', help='start a frontend service'))
    set_indexer_service_parser(sp.add_parser('index', help='start an indexer service'))
    set_loadable_service_parser(sp.add_parser('encode', help='start an encoder service'))
    set_router_service_parser(sp.add_parser('route', help='start a router service'))
    set_preprocessor_service_parser(sp.add_parser('preprocess', help='start a preprocessor service'))
    set_grpc_service_parser(sp.add_parser('grpc', help='start a general purpose grpc service'))

    # clients
    set_http_service_parser(sp.add_parser('client_http', help='start a http service'))
    set_cli_client_parser(sp.add_parser('client_cli', help='start a grpc client'))

    # others
    set_composer_flask_parser(sp.add_parser('compose', help='start a GNES Board and visualize YAML config'))
    return parser



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


# A key-value map for Class to the (module)file it located in
from ..base import register_all_class

_cls2file_map = {
    'BertEncoder': 'text.bert',
    'BertEncoderWithServer': 'text.bert',
    'BertEncoderServer': 'text.bert',
    'ElmoEncoder': 'text.elmo',
    'FlairEncoder': 'text.flair',
    'GPTEncoder': 'text.gpt',
    'GPT2Encoder': 'text.gpt',
    'PCALocalEncoder': 'numeric.pca',
    'PQEncoder': 'numeric.pq',
    'TFPQEncoder': 'numeric.tf_pq',
    'Word2VecEncoder': 'text.w2v',
    'BaseEncoder': 'base',
    'BaseBinaryEncoder': 'base',
    'BaseTextEncoder': 'base',
    'BaseVideoEncoder': 'base',
    'BaseNumericEncoder': 'base',
    'BaseAudioEncoder': 'base',
    'PipelineEncoder': 'base',
    'HashEncoder': 'numeric.hash',
    'BasePytorchEncoder': 'image.base',
    'TFInceptionEncoder': 'image.inception',
    'CVAEEncoder': 'image.cvae',
    'IncepMixtureEncoder': 'video.incep_mixture',
    'VladEncoder': 'numeric.vlad',
    'MfccEncoder': 'audio.mfcc'
}

register_all_class(_cls2file_map, 'encoder')



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import math

import tensorflow as tf
import tensorflow.contrib.slim as slim


class NetFV():
    def __init__(self, feature_size,
                 cluster_size,
                 vocab_size,
                 method='netvlad',
                 use_length=False,
                 input_size=None,
                 use_2nd_label=False,
                 vocab_size_2=None,
                 add_batch_norm=True,
                 is_training=False,
                 use_weights=True,
                 save_dir=None,
                 multitask_method=None,
                 l2_penalty=1e-6):
        if input_size == None:
            self.input_size = feature_size
        else:
            self.input_size = input_size
        self.feature_size = feature_size
        self.use_length = use_length
        self.is_training = is_training
        self.vocab_size = vocab_size
        self.use_2nd_label = use_2nd_label
        self.vocab_size_2 = vocab_size_2
        self.add_batch_norm = add_batch_norm
        self.cluster_size = cluster_size
        self.use_weights = use_weights
        self.l2_penalty = l2_penalty
        self.method = method
        self.multitask_method = multitask_method
        self.build_model()
        self.build_loss()

    @staticmethod
    def rand_init(feature_size):
        return tf.random_normal_initializer(stddev=1 / math.sqrt(feature_size))

    def build_model(self):
        self.feeds = tf.placeholder(tf.float32, [None, None, self.input_size])
        self.feeds_length = tf.placeholder(tf.int32, [None])

        self.inputs = tf.layers.dense(self.feeds, self.feature_size)

        self.weights = tf.placeholder(tf.float32, [None, self.vocab_size])
        self.max_frames = tf.shape(self.inputs)[1]
        self.seq_length = tf.cast(tf.sequence_mask(self.feeds_length,
                                                   self.max_frames), tf.float32)
        if self.method == 'fvnet':
            self.build_fvnet()
        elif self.method == 'netvlad':
            self.build_netvlad()
        elif self.method == 'pooling':
            self.build_pooling()

    def build_pooling(self):
        self.repre = tf.layers.dense(self.inputs, self.feature_size)
        self.repre = tf.reduce_max(self.repre, axis=1)

    def build_fvnet(self):
        reshaped_input = tf.reshape(self.inputs, [-1, self.feature_size])
        cluster_weights = tf.get_variable("cluster_weights",
                                          [self.feature_size, self.cluster_size],
                                          initializer=NetFV.rand_init(self.feature_size))

        covar_weights = tf.get_variable("covar_weights",
                                        [self.feature_size, self.cluster_size],
                                        initializer=NetFV.rand_init(self.feature_size))

        covar_weights = tf.square(covar_weights)
        eps = tf.constant([1e-6])
        covar_weights = tf.add(covar_weights, eps)

        tf.summary.histogram("cluster_weights", cluster_weights)
        activation = tf.matmul(reshaped_input, cluster_weights)
        if self.add_batch_norm:
            activation = slim.batch_norm(activation,
                                         center=True,
                                         scale=True,
                                         is_training=self.is_training,
                                         scope="cluster_bn")
        else:
            cluster_biases = tf.get_variable("cluster_biases",
                                             [self.cluster_size],
                                             initializer=NetFV.rand_init(self.feature_size))
            tf.summary.histogram("cluster_biases", cluster_biases)
            activation += cluster_biases

        activation = tf.nn.softmax(activation)
        tf.summary.histogram("cluster_output", activation)

        activation = tf.reshape(activation, [-1, self.max_frames, self.cluster_size])

        a_sum = tf.reduce_sum(activation, -2, keepdims=True)

        cluster_weights2 = tf.scalar_mul(0.01, cluster_weights)

        a = tf.multiply(a_sum, cluster_weights2)

        activation = tf.transpose(activation, perm=[0, 2, 1])

        reshaped_input = tf.reshape(reshaped_input,
                                    [-1, self.max_frames, self.feature_size])
        fv1 = tf.matmul(activation, reshaped_input)

        fv1 = tf.transpose(fv1, perm=[0, 2, 1])

        # computing second order FV
        a2 = tf.multiply(a_sum, tf.square(cluster_weights2))

        b2 = tf.multiply(fv1, cluster_weights2)
        fv2 = tf.matmul(activation, tf.square(reshaped_input))

        fv2 = tf.transpose(fv2, perm=[0, 2, 1])
        fv2 = tf.add_n([a2, fv2, tf.scalar_mul(-2, b2)])

        fv2 = tf.divide(fv2, tf.square(covar_weights))
        fv2 = tf.subtract(fv2, a_sum)

        fv2 = tf.reshape(fv2, [-1, self.cluster_size * self.feature_size])
        fv2 = tf.nn.l2_normalize(fv2, 1)
        fv2 = tf.reshape(fv2, [-1, self.cluster_size * self.feature_size])
        fv2 = tf.nn.l2_normalize(fv2, 1)

        fv1 = tf.subtract(fv1, a)
        fv1 = tf.divide(fv1, covar_weights)
        fv1 = tf.nn.l2_normalize(fv1, 1)
        fv1 = tf.reshape(fv1, [-1, self.cluster_size * self.feature_size])
        fv1 = tf.nn.l2_normalize(fv1, 1)

        self.repre = tf.concat([fv1, fv2], 1)
        self.repre = tf.layers.dense(self.repre, self.feature_size)

    def build_netvlad(self):
        reshaped_input = tf.reshape(self.inputs, [-1, self.feature_size])
        cluster_weights = tf.get_variable("cluster_weights",
                                          [self.feature_size, self.cluster_size],
                                          initializer=NetFV.rand_init(self.feature_size))
        activation = tf.matmul(reshaped_input, cluster_weights)
        if self.add_batch_norm:
            activation = slim.batch_norm(activation,
                                         center=True,
                                         scale=True,
                                         is_training=self.is_training,
                                         scope="cluster_bn")
        else:
            cluster_biases = tf.get_variable("cluster_biases",
                                             [self.cluster_size],
                                             initializer=NetFV.rand_init(self.feature_size))
            activation += cluster_biases
        activation = tf.nn.softmax(activation)
        activation = tf.reshape(activation, [-1, self.max_frames, self.cluster_size])
        if self.use_length:
            activation *= tf.reshape(self.seq_length, [-1, self.max_frames, 1])

        a_sum = tf.reduce_sum(activation, -2, keep_dims=True)

        if self.use_length:
            a_sum = a_sum / tf.cast(tf.reshape(self.feeds_length, [-1, 1, 1]), tf.float32)

        cluster_weights2 = tf.get_variable("cluster_weights2",
                                           [1, self.feature_size, self.cluster_size],
                                           initializer=NetFV.rand_init(self.feature_size))

        a = tf.multiply(a_sum, cluster_weights2)
        activation = tf.transpose(activation, perm=[0, 2, 1])

        reshaped_input = tf.reshape(reshaped_input,
                                    [-1, self.max_frames, self.feature_size])
        vlad = tf.matmul(activation, reshaped_input)
        if self.use_length:
            vlad = vlad / tf.cast(tf.reshape(self.feeds_length, [-1, 1, 1]), tf.float32)
        vlad = tf.transpose(vlad, perm=[0, 2, 1])
        vlad = tf.subtract(vlad, a)

        vlad = tf.nn.l2_normalize(vlad, 1)

        vlad = tf.reshape(vlad, [-1, self.cluster_size * self.feature_size])
        vlad = tf.nn.l2_normalize(vlad, 1)
        self.repre = vlad

    def build_loss(self):
        self.probabilities = tf.layers.dense(self.repre,
                                             self.vocab_size,
                                             activation=tf.nn.tanh)
        self.probabilities = tf.layers.dense(self.probabilities, self.vocab_size)
        self.probabilities = tf.nn.softmax(self.probabilities)

        self.label = tf.placeholder(tf.int32, [None, self.vocab_size])
        logits = tf.cast(self.label, tf.float32)
        if self.use_weights:
            logits = logits * self.weights
        self.loss = - tf.log(tf.reduce_sum(logits * self.probabilities, axis=1) + 1e-9)
        self.loss = tf.reduce_mean(self.loss)
        self.pred = tf.argmax(self.probabilities, 1)
        self.avg_diff = tf.cast(tf.equal(tf.argmax(self.label, 1), self.pred), tf.float32)
        self.avg_diff = tf.reduce_mean(self.avg_diff)

        # add 2nd layer labels
        if self.use_2nd_label:
            self.label_2 = tf.placeholder(tf.int32, [None, self.vocab_size_2])
            logits2 = tf.cast(self.label_2, tf.float32)

            if self.multitask_method is None:
                self.probabilities2 = tf.layers.dense(self.repre,
                                                      self.vocab_size_2,
                                                      activation=tf.nn.tanh)
                self.probabilities2 = tf.layers.dense(self.probabilities2, self.vocab_size_2)
                self.probabilities2 = tf.nn.softmax(self.probabilities2)

            elif self.multitask_method == 'Attention':
                self.x = tf.get_variable('emb',
                                         shape=[self.vocab_size, self.feature_size],
                                         dtype=tf.float32,
                                         initializer=NetFV.rand_init(self.feature_size))
                self.emb_label = tf.matmul(self.probabilities, self.x)
                self.emb_concat = tf.concat([self.emb_label, self.repre], axis=1)
                self.probabilities2 = tf.layers.dense(self.emb_concat,
                                                      self.vocab_size_2,
                                                      activation=tf.nn.tanh)
                self.probabilities2 = tf.layers.dense(self.probabilities2,
                                                      self.vocab_size_2)
                self.probabilities2 = tf.nn.softmax(self.probabilities2)

            self.loss += tf.reduce_mean(-tf.log(
                tf.reduce_sum(logits2 * self.probabilities2, axis=1) + 1e-9))
            self.pred2 = tf.argmax(self.probabilities2, 1)
            self.avg_diff2 = tf.cast(tf.equal(tf.argmax(self.label_2, 1), self.pred2), tf.float32)
            self.avg_diff2 = tf.reduce_mean(self.avg_diff2)

        self.optimizer = tf.train.AdamOptimizer(learning_rate=0.0005,
                                                epsilon=1e-08,
                                                name='adam')
        self.train_op = slim.learning.create_train_op(self.loss, self.optimizer)
        self.eval_res = {'loss': self.loss, 'avg_diff': self.avg_diff}
        if self.use_2nd_label:
            self.eval_res['avg_diff2'] = self.avg_diff2



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List

import numpy as np
from PIL import Image

from ..base import BaseVideoEncoder
from ...helper import batching, get_first_available_gpu


class IncepMixtureEncoder(BaseVideoEncoder):
    batch_size = 64

    def __init__(self, model_dir_inception: str,
                 model_dir_mixture: str,
                 select_layer: str = 'PreLogitsFlatten',
                 use_cuda: bool = False,
                 feature_size: int = 300,
                 vocab_size: int = 28,
                 cluster_size: int = 256,
                 method: str = 'fvnet',
                 input_size: int = 1536,
                 vocab_size_2: int = 174,
                 max_frames: int = 30,
                 multitask_method: str = 'Attention',
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_dir_inception = model_dir_inception
        self.model_dir_mixture = model_dir_mixture
        self.select_layer = select_layer
        self.use_cuda = use_cuda
        self.cluster_size = cluster_size
        self.feature_size = feature_size
        self.vocab_size = vocab_size
        self.method = method
        self.input_size = input_size
        self.multitask_method = multitask_method
        self.inception_size_x = 299
        self.inception_size_y = 299
        self.max_frames = max_frames
        self.vocab_size_2 = vocab_size_2

    def post_init(self):
        import tensorflow as tf
        from ..image.inception_cores.inception_v4 import inception_v4
        from ..image.inception_cores.inception_utils import inception_arg_scope
        from .mixture_core.model import NetFV
        import os
        os.environ['CUDA_VISIBLE_DEVICES'] = str(get_first_available_gpu())

        g = tf.Graph()
        with g.as_default():
            arg_scope = inception_arg_scope()
            inception_v4.default_image_size = self.inception_size_x
            self.inputs = tf.placeholder(tf.float32, (None,
                                                      self.inception_size_x,
                                                      self.inception_size_y, 3))

            with tf.contrib.slim.arg_scope(arg_scope):
                self.logits, self.end_points = inception_v4(self.inputs,
                                                            is_training=False,
                                                            dropout_keep_prob=1.0)

            config = tf.ConfigProto(log_device_placement=False)
            if self.use_cuda:
                config.gpu_options.allow_growth = True
            self.sess = tf.Session(config=config)
            self.saver = tf.train.Saver()
            self.saver.restore(self.sess, self.model_dir_inception)

        g2 = tf.Graph()
        with g2.as_default():
            config = tf.ConfigProto(log_device_placement=False)
            if self.use_cuda:
                config.gpu_options.allow_growth = True
            self.sess2 = tf.Session(config=config)
            self.mix_model = NetFV(feature_size=self.feature_size,
                                   cluster_size=self.cluster_size,
                                   vocab_size=self.vocab_size,
                                   input_size=self.input_size,
                                   use_2nd_label=True,
                                   vocab_size_2=self.vocab_size_2,
                                   multitask_method=self.multitask_method,
                                   method=self.method,
                                   is_training=False)
            saver = tf.train.Saver(max_to_keep=1)
            self.sess2.run(tf.global_variables_initializer())
            saver.restore(self.sess2, self.model_dir_mixture)

    def encode(self, data: List['np.ndarray'], *args, **kwargs) -> np.ndarray:
        v_len = [len(v) for v in data]
        pos_start = [0] + [sum(v_len[:i + 1]) for i in range(len(v_len) - 1)]
        pos_end = [sum(v_len[:i + 1]) for i in range(len(v_len))]
        max_len = min(max(v_len), self.max_frames)

        img = [im for v in data for im in v]
        img = [(np.array(Image.fromarray(im).resize((self.inception_size_x,
                                                     self.inception_size_y)), dtype=np.float32) * 2 / 255. - 1.) for im
               in img]

        @batching(concat_axis=None)
        def _encode1(self, data):
            _, end_points_ = self.sess.run((self.logits, self.end_points),
                                           feed_dict={self.inputs: data})
            return end_points_[self.select_layer]

        v = [_ for vi in _encode1(self, img) for _ in vi]

        v_input = [v[s:e] for s, e in zip(pos_start, pos_end)]
        v_input = [(vi + [[0.0] * self.input_size] * (max_len - len(vi)))[:max_len] for vi in v_input]
        v_input = [np.array(vi, dtype=np.float32) for vi in v_input]

        @batching
        def _encode2(self, data):
            return self.sess2.run(self.mix_model.repre,
                                  feed_dict={self.mix_model.feeds: data})

        return _encode2(self, v_input).astype(np.float32)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import asyncio
import json
from concurrent.futures import ThreadPoolExecutor

import grpc
from google.protobuf.json_format import MessageToJson

from ..helper import set_logger
from ..proto import gnes_pb2_grpc, RequestGenerator


class HttpClient:
    def __init__(self, args=None):
        self.args = args
        self.logger = set_logger(self.__class__.__name__, self.args.verbose)

    def start(self):
        from aiohttp import web
        loop = asyncio.get_event_loop()
        executor = ThreadPoolExecutor(max_workers=self.args.max_workers)

        async def general_handler(request, parser, *args, **kwargs):
            try:
                data = dict()
                # # Option 1: uploading streaming chunk data
                # data = b""
                # async for chunk in request.content.iter_any():
                #     data += chunk
                # self.logger.info("received %d content" % len(data))

                # Option 2: uploading via Multipart-Encoded File
                post_data = await request.post()
                if 'query' in post_data.keys():
                    _file = post_data.get('query')
                    self.logger.info("query request from input file: %s" % _file.filename)
                    data['query'] = _file.file.read()
                elif 'docs' in post_data.keys():
                    files = post_data.getall("docs")
                    self.logger.info("index request from input files: %d files" % len(files))
                    data['docs'] = [_file.file.read() for _file in files]

                self.logger.info('data received, beigin processing')
                resp = await loop.run_in_executor(
                    executor,
                    stub_call,
                    parser([d for d in data.get('docs')] if hasattr(data, 'docs')
                           else data.get('query'), *args, **kwargs))
                self.logger.info('handling finished, will send to user')
                return web.Response(body=json.dumps({'result': resp, 'meta': None}, ensure_ascii=False),
                                    status=200,
                                    content_type='application/json')
            except Exception as ex:
                return web.Response(body=json.dumps({'message': str(ex), 'type': type(ex)}),
                                    status=400,
                                    content_type='application/json')

        async def init(loop):
            # persistant connection or non-persistant connection
            handler_args = {'tcp_keepalive': False, 'keepalive_timeout': 25}
            app = web.Application(loop=loop,
                                  client_max_size=10 ** 10,
                                  handler_args=handler_args)
            app.router.add_route('post', '/train',
                                 lambda x: general_handler(x, RequestGenerator.train, batch_size=self.args.batch_size))
            app.router.add_route('post', '/index',
                                 lambda x: general_handler(x, RequestGenerator.index, batch_size=self.args.batch_size))
            app.router.add_route('post', '/query',
                                 lambda x: general_handler(x, RequestGenerator.query, top_k=self.args.top_k))
            srv = await loop.create_server(app.make_handler(),
                                           self.args.http_host,
                                           self.args.http_port)
            self.logger.info('http server listens to %s:%d ...' % (self.args.http_host, self.args.http_port))
            return srv

        def stub_call(req):
            res_f = list(stub.StreamCall(req))[-1]
            return json.loads(MessageToJson(res_f))

        with grpc.insecure_channel(
                '%s:%s' % (self.args.grpc_host, self.args.grpc_port),
                options=[('grpc.max_send_message_length', self.args.max_message_size * 1024 * 1024),
                         ('grpc.max_receive_message_length', self.args.max_message_size * 1024 * 1024),
                         ('grpc.keepalive_timeout_ms', 100 * 1000)]) as channel:
            stub = gnes_pb2_grpc.GnesRPCStub(channel)
            loop.run_until_complete(init(loop))
            loop.run_forever()



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import numpy as np

from .base import BaseAudioPreprocessor
from ..helper import get_video_length_from_raw, get_audio
from ...proto import array2blob


class AudioVanilla(BaseAudioPreprocessor):

    def __init__(self, audio_interval: int,
                 sample_rate: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.audio_interval = audio_interval
        self.sample_rate = sample_rate

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        super().apply(doc)

        if doc.raw_bytes:
            audio = get_audio(doc.raw_bytes, self.sample_rate,
                              self.audio_interval, get_video_length_from_raw(doc.raw_bytes))
            if len(audio) >= 1:
                for ci, chunks in enumerate(audio):
                    c = doc.chunks.add()
                    c.doc_id = doc.doc_id
                    c.blob.CopyFrom(array2blob(np.array(chunks, dtype=np.float32)))
                    c.offset_1d = ci
                    c.weight = 1 / len(audio)
            else:
                self.logger.info('bad document: no audio extracted')
        else:
            self.logger.error('bad document: "raw_bytes" is empty!')



import os
import sys
import unittest.mock

dirname = os.path.dirname(__file__)
module_path = os.path.join(dirname, 'contrib', 'dummy_contrib.py')
cls_name = 'FooContribEncoder'

@unittest.SkipTest
class TestContribModule(unittest.TestCase):
    def setUp(self):
        self.yaml_path = os.path.join(os.path.dirname(__file__),
                                      'contrib', 'dummy.yml')

    def tearDown(self):
        # reload gnes module on every unit test
        for mod in list(sys.modules.keys()):
            if mod.startswith('gnes.'):
                del (sys.modules[mod])

    @unittest.mock.patch.dict(os.environ, {'GNES_CONTRIB_MODULE': '%s:%s' % (cls_name, module_path)})
    def test_load_contrib(self):

        from gnes.encoder.base import BaseEncoder, BaseTextEncoder
        a = BaseEncoder.load_yaml(self.yaml_path)
        self.assertIsInstance(a, BaseTextEncoder)
        self.assertEqual(a.encode([]), 'hello 531')

    @unittest.mock.patch.dict(os.environ, {'GNES_CONTRIB_MODULE': '%s:%s' % ('blah', module_path)})
    def test_bad_name(self):
        try:
            from gnes.encoder.base import BaseEncoder
        except AttributeError:
            pass

    @unittest.mock.patch.dict(os.environ, {'GNES_CONTRIB_MODULE': '%s:%s' % (cls_name, 'blah')})
    def test_bad_path(self):
        try:
            from gnes.encoder.base import BaseEncoder
        except AttributeError:
            pass



import copy
import os
import unittest
import zipfile

from gnes.encoder.image.onnx import BaseONNXImageEncoder
from gnes.preprocessor.base import BaseUnaryPreprocessor, PipelinePreprocessor
from gnes.preprocessor.image.resize import ResizeChunkPreprocessor
from gnes.preprocessor.image.sliding_window import VanillaSlidingPreprocessor
from gnes.proto import gnes_pb2, blob2array

def img_process_for_test(dirname):
    zipfile_ = zipfile.ZipFile(os.path.join(dirname, 'imgs/test.zip'))
    all_bytes = [zipfile_.open(v).read() for v in zipfile_.namelist()]
    test_img = []
    for raw_bytes in all_bytes:
        d = gnes_pb2.Document()
        d.raw_bytes = raw_bytes
        test_img.append(d)

    test_img_all_preprocessor = []
    pipline_prep1 = PipelinePreprocessor()
    pipline_prep1.component = lambda: [BaseUnaryPreprocessor(doc_type=gnes_pb2.Document.IMAGE),
                                       ResizeChunkPreprocessor()]
    for preprocessor in [pipline_prep1,
                         VanillaSlidingPreprocessor()]:
        test_img_copy = copy.deepcopy(test_img)
        for img in test_img_copy:
            preprocessor.apply(img)
        test_img_all_preprocessor.append([blob2array(chunk.blob)
                                          for img in test_img_copy for chunk in img.chunks])
    return test_img_all_preprocessor

class TestONNXImageEncoder(unittest.TestCase):

    def setUp(self):
        dirname = os.path.dirname(__file__)
        self.dump_path = os.path.join(dirname, 'model.bin')
        self.test_img = img_process_for_test(dirname)
        self.vgg_yaml = os.path.join(dirname, 'yaml', 'onnx-vgg-encoder.yml')
        self.res_yaml = os.path.join(dirname, 'yaml', 'onnx-resnet-encoder.yml')
        self.inception_yaml = os.path.join(dirname, 'yaml', 'onnx-inception-encoder.yml')
        self.mobilenet_yaml = os.path.join(dirname, 'yaml', 'onnx-mobilenet-encoder.yml')

    def test_vgg_encoding(self):
        self.encoder = BaseONNXImageEncoder.load_yaml(self.vgg_yaml)
        for test_img in self.test_img:
            vec = self.encoder.encode(test_img)
            print("the length of data now is:", len(test_img))
            self.assertEqual(vec.shape[0], len(test_img))
            self.assertEqual(vec.shape[1], 1000)

    def test_resnet_encoding(self):
        self.encoder = BaseONNXImageEncoder.load_yaml(self.res_yaml)
        for test_img in self.test_img:
            vec = self.encoder.encode(test_img)
            print("the length of data now is:", len(test_img))
            self.assertEqual(vec.shape[0], len(test_img))
            self.assertEqual(vec.shape[1], 1000)

    def test_inception_encoding(self):
        self.encoder = BaseONNXImageEncoder.load_yaml(self.inception_yaml)
        for test_img in self.test_img:
            vec = self.encoder.encode(test_img)
            print("the length of data now is:", len(test_img))
            self.assertEqual(vec.shape[0], len(test_img))
            self.assertEqual(vec.shape[1], 1000)

    def test_mobilenet_encoding(self):
        self.encoder = BaseONNXImageEncoder.load_yaml(self.mobilenet_yaml)
        for test_img in self.test_img:
            vec = self.encoder.encode(test_img)
            print("the length of data now is:", len(test_img))
            self.assertEqual(vec.shape[0], len(test_img))
            self.assertEqual(vec.shape[1], 1000)

    def test_dump_load(self):
        self.encoder = BaseONNXImageEncoder.load_yaml(self.inception_yaml)

        self.encoder.dump(self.dump_path)

        vgg_encoder2 = BaseONNXImageEncoder.load(self.dump_path)

        for test_img in self.test_img:
            vec = vgg_encoder2.encode(test_img)
            self.assertEqual(vec.shape[0], len(test_img))
            self.assertEqual(vec.shape[1], 1000)

    def tearDown(self):
        if os.path.exists(self.dump_path):
            os.remove(self.dump_path)



import os
import unittest
import numpy as np
from gnes.indexer.vector.hbindexer import HBIndexer
import shutil

class TestMHIndexer(unittest.TestCase):

    def setUp(self):
        self.num_clusters = 100
        self.num_bytes = 16
        self.n_idx = 1
        self.n = 100

        self.test_label = [(_, 1) for _ in range(self.n)]
        t = np.random.randint(0, 100, size=[self.n, self.n_idx+self.num_bytes])
        self.test_data = t.astype(np.uint32)
        self.weights = [1.] * len(self.test_label)
        self.data_path = 'test_path'
        self.dump_path = './dump.bin'

        self.query = self.test_data

    def tearDown(self):
        if os.path.exists(self.dump_path):
            os.remove(self.dump_path)

    def test_add_query(self):
        m = HBIndexer(self.num_clusters, self.num_bytes, self.n_idx, self.data_path)
        m.add(self.test_label, self.test_data, self.weights)
        res = m.query(self.query, 1)
        self.assertEqual(len(res), self.n)
        s = sum([1 for i in range(self.n) if i in [_[0][0] for _ in res[i]]])
        self.assertEqual(s, self.n)
        m.close()

    def test_dump_load(self):
        m = HBIndexer(self.num_clusters, self.num_bytes, self.n_idx, self.data_path)
        m.add(self.test_label, self.test_data, self.weights)
        m.dump(self.dump_path)
        m.close()
        self.assertTrue(os.path.exists(self.dump_path))
        m2 = HBIndexer.load(self.dump_path)
        res = m2.query(self.query, 1)
        s = sum([1 for i in range(self.n) if i in [_[0][0] for _ in res[i]]])
        self.assertEqual(s, self.n)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.



import os
from typing import List, Tuple, Any

import numpy as np

from .cython import IndexCore
from ...base import BaseVectorIndexer


class HBIndexer(BaseVectorIndexer):

    def __init__(self,
                 num_clusters: int = 100,
                 num_bytes: int = 8,
                 n_idx: int = 1,
                 data_path: str = './hbindexer_data',
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.n_bytes = num_bytes
        self.n_clusters = num_clusters
        self.n_idx = n_idx
        self.data_path = data_path
        self._weight_norm = 2 ** 16 - 1
        if self.n_idx <= 0:
            raise ValueError('There should be at least 1 clustering slot')

    def post_init(self):
        self.hbindexer = IndexCore(self.n_clusters, self.n_bytes, self.n_idx)
        try:
            if not os.path.exists(self.data_path):
                raise FileNotFoundError('"data_path" is not exist')
            if os.path.isdir(self.data_path):
                raise IsADirectoryError('"data_path" must be a file path, not a directory')
            self.hbindexer.load(self.data_path)
        except (FileNotFoundError, IsADirectoryError):
            self.logger.warning('fail to load model from %s, will create an empty one' % self.data_path)

    def add(self, keys: List[Tuple[int, Any]], vectors: np.ndarray, weights: List[float], *args, **kwargs):
        if len(vectors) != len(keys):
            raise ValueError("vectors length should be equal to doc_ids")

        if vectors.dtype != np.uint32:
            raise ValueError("vectors should be ndarray of uint32")

        n = len(keys)
        keys, offsets = zip(*keys)
        keys = np.array(keys, dtype=np.uint32).tobytes()
        offsets = np.array(offsets, dtype=np.uint16).tobytes()
        weights = np.array(weights * self._weight_norm, dtype=np.uint16).tobytes()
        clusters = vectors[:, :self.n_idx].tobytes()
        vectors = vectors[:, self.n_idx:].astype(np.uint8).tobytes()
        self.hbindexer.index_trie(vectors, clusters, keys, offsets, weights, n)

    def query(self,
              vectors: np.ndarray,
              top_k: int,
              *args,
              **kwargs) -> List[List[Tuple]]:

        if vectors.dtype != np.uint32:
            raise ValueError("vectors should be ndarray of uint32")

        # num_rows = int(len(keys) / self.num_bytes)
        n = vectors.shape[0]
        clusters = vectors[:, :self.n_idx].tobytes()
        vectors = vectors[:, self.n_idx:].astype(np.uint8).tobytes()

        result = [{} for _ in range(n)]

        doc_ids, offsets, weights, dists, q_idx = self.hbindexer.query(
            vectors, clusters, n, top_k * self.n_idx)
        for (i, o, w, d, q) in zip(doc_ids, offsets, weights, dists, q_idx):
            result[q][(i, o, w / self._weight_norm)] = self.normalize_score(d)

        return [sorted(ret.items(), key=lambda x: -x[1])[:top_k] for ret in result]

    def normalize_score(self, distance: int, *args, **kwargs) -> float:
        return 1. - distance / self.n_bytes * 8

    def __getstate__(self):
        self.hbindexer.save(self.data_path)
        d = super().__getstate__()
        return d



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import gnes_pb2 as gnes__pb2


class GnesRPCStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.Train = channel.unary_unary(
        '/gnes.GnesRPC/Train',
        request_serializer=gnes__pb2.Request.SerializeToString,
        response_deserializer=gnes__pb2.Response.FromString,
        )
    self.Index = channel.unary_unary(
        '/gnes.GnesRPC/Index',
        request_serializer=gnes__pb2.Request.SerializeToString,
        response_deserializer=gnes__pb2.Response.FromString,
        )
    self.Query = channel.unary_unary(
        '/gnes.GnesRPC/Query',
        request_serializer=gnes__pb2.Request.SerializeToString,
        response_deserializer=gnes__pb2.Response.FromString,
        )
    self.Call = channel.unary_unary(
        '/gnes.GnesRPC/Call',
        request_serializer=gnes__pb2.Request.SerializeToString,
        response_deserializer=gnes__pb2.Response.FromString,
        )
    self.StreamCall = channel.stream_stream(
        '/gnes.GnesRPC/StreamCall',
        request_serializer=gnes__pb2.Request.SerializeToString,
        response_deserializer=gnes__pb2.Response.FromString,
        )


class GnesRPCServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def Train(self, request, context):
    """option (rpc_core.method_no_deadline) = true;
    option (rpc_core.service_default_deadline_ms) = 5000;
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Index(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Query(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Call(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def StreamCall(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_GnesRPCServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'Train': grpc.unary_unary_rpc_method_handler(
          servicer.Train,
          request_deserializer=gnes__pb2.Request.FromString,
          response_serializer=gnes__pb2.Response.SerializeToString,
      ),
      'Index': grpc.unary_unary_rpc_method_handler(
          servicer.Index,
          request_deserializer=gnes__pb2.Request.FromString,
          response_serializer=gnes__pb2.Response.SerializeToString,
      ),
      'Query': grpc.unary_unary_rpc_method_handler(
          servicer.Query,
          request_deserializer=gnes__pb2.Request.FromString,
          response_serializer=gnes__pb2.Response.SerializeToString,
      ),
      'Call': grpc.unary_unary_rpc_method_handler(
          servicer.Call,
          request_deserializer=gnes__pb2.Request.FromString,
          response_serializer=gnes__pb2.Response.SerializeToString,
      ),
      'StreamCall': grpc.stream_stream_rpc_method_handler(
          servicer.StreamCall,
          request_deserializer=gnes__pb2.Request.FromString,
          response_serializer=gnes__pb2.Response.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'gnes.GnesRPC', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



import os
import unittest
import zipfile

from gnes.cli.parser import set_preprocessor_service_parser, _set_client_parser
from gnes.client.base import ZmqClient
from gnes.proto import gnes_pb2, RequestGenerator, blob2array
from gnes.service.preprocessor import PreprocessorService


class TestProto(unittest.TestCase):

    def setUp(self):
        self.dirname = os.path.dirname(__file__)
        self.unary_img_pre_yaml = os.path.join(self.dirname, 'yaml', 'base-unary-image-prep.yml')
        self.slidingwindow_img_pre_yaml = os.path.join(self.dirname, 'yaml', 'base-vanilla_sldwin-image-prep.yml')
        self.segmentation_img_pre_yaml = os.path.join(self.dirname, 'yaml', 'base-segmentation-image-prep.yml')
        self.resize_img_pre_yaml = os.path.join(self.dirname, 'yaml', 'resize-image-prep.yml')

    def test_unary_preprocessor_service_empty(self):
        args = set_preprocessor_service_parser().parse_args([
            '--yaml_path', self.unary_img_pre_yaml
        ])
        with PreprocessorService(args):
            pass

    def test_slidingwindow_preprocessor_service_empty(self):
        args = set_preprocessor_service_parser().parse_args([
            '--yaml_path', self.slidingwindow_img_pre_yaml
        ])
        with PreprocessorService(args):
            pass

    def test_segmentation_preprocessor_service_empty(self):
        args = set_preprocessor_service_parser().parse_args([
            '--yaml_path', self.segmentation_img_pre_yaml
        ])
        with PreprocessorService(args):
            pass

    def test_unary_preprocessor_service_echo(self):
        args = set_preprocessor_service_parser().parse_args([
            '--yaml_path', self.unary_img_pre_yaml
        ])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in)
        ])
        with PreprocessorService(args), ZmqClient(c_args) as client:
            msg = gnes_pb2.Message()
            msg.request.index.docs.extend([gnes_pb2.Document() for _ in range(5)])
            client.send_message(msg)
            r = client.recv_message()
            # print(r)
            msg.request.train.docs.extend([gnes_pb2.Document() for _ in range(5)])
            client.send_message(msg)
            r = client.recv_message()
            # print(r)

    def test_slidingwindow_preprocessor_service_echo(self):
        args = set_preprocessor_service_parser().parse_args([
            '--yaml_path', self.slidingwindow_img_pre_yaml
        ])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in)
        ])
        with PreprocessorService(args), ZmqClient(c_args) as client:
            msg = gnes_pb2.Message()
            msg.request.index.docs.extend([gnes_pb2.Document() for _ in range(5)])
            client.send_message(msg)
            r = client.recv_message()
            # print(r)
            msg.request.train.docs.extend([gnes_pb2.Document() for _ in range(5)])
            client.send_message(msg)
            r = client.recv_message()
            # print(r)

    def test_segmentation_preprocessor_service_echo(self):
        args = set_preprocessor_service_parser().parse_args([
            '--yaml_path', self.segmentation_img_pre_yaml
        ])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in)
        ])
        with PreprocessorService(args), ZmqClient(c_args) as client:
            msg = gnes_pb2.Message()
            msg.request.index.docs.extend([gnes_pb2.Document() for _ in range(5)])
            client.send_message(msg)
            r = client.recv_message()
            # print(r)
            msg.request.train.docs.extend([gnes_pb2.Document() for _ in range(5)])
            client.send_message(msg)
            r = client.recv_message()
            # print(r)

    def test_unary_preprocessor_service_realdata(self):
        args = set_preprocessor_service_parser().parse_args([
            '--yaml_path', self.unary_img_pre_yaml
        ])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in)
        ])
        all_zips = zipfile.ZipFile(os.path.join(self.dirname, 'imgs/test.zip'))
        all_bytes = [all_zips.open(v).read() for v in all_zips.namelist()]

        with PreprocessorService(args), ZmqClient(c_args) as client:
            for req in RequestGenerator.index(all_bytes):
                msg = gnes_pb2.Message()
                msg.request.index.CopyFrom(req.index)
                client.send_message(msg)
                r = client.recv_message()
                self.assertEqual(r.envelope.routes[0].service, 'PreprocessorService:BaseUnaryPreprocessor')
                for d in r.request.index.docs:
                    self.assertEqual(len(d.chunks), 1)
                    self.assertEqual(len(blob2array(d.chunks[0].blob).shape), 3)
                    self.assertEqual(blob2array(d.chunks[0].blob).shape[-1], 3)

    def test_resize_preprocessor_service_realdata(self):
        args = set_preprocessor_service_parser().parse_args([
            '--yaml_path', self.resize_img_pre_yaml
        ])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in)
        ])
        all_zips = zipfile.ZipFile(os.path.join(self.dirname, 'imgs/test.zip'))
        all_bytes = [all_zips.open(v).read() for v in all_zips.namelist()]

        with PreprocessorService(args), ZmqClient(c_args) as client:
            for req in RequestGenerator.index(all_bytes):
                msg = gnes_pb2.Message()
                msg.request.index.CopyFrom(req.index)
                client.send_message(msg)
                r = client.recv_message()
                self.assertEqual(r.envelope.routes[0].service, 'PreprocessorService:PipelinePreprocessor')
                for d in r.request.index.docs:
                    self.assertEqual(len(d.chunks), 1)
                    self.assertEqual(len(blob2array(d.chunks[0].blob).shape), 3)
                    self.assertEqual(blob2array(d.chunks[0].blob).shape[-1], 3)
                    self.assertEqual(blob2array(d.chunks[0].blob).shape[0], 224)
                    self.assertEqual(blob2array(d.chunks[0].blob).shape[1], 224)

    def test_slidingwindow_preprocessor_service_realdata(self):
        args = set_preprocessor_service_parser().parse_args([
            '--yaml_path', self.slidingwindow_img_pre_yaml
        ])

        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in)
        ])
        all_zips = zipfile.ZipFile(os.path.join(self.dirname, 'imgs/test.zip'))
        all_bytes = [all_zips.open(v).read() for v in all_zips.namelist()]

        with PreprocessorService(args), ZmqClient(c_args) as client:
            for req in RequestGenerator.index(all_bytes):
                msg = gnes_pb2.Message()
                msg.request.index.CopyFrom(req.index)
                client.send_message(msg)
                r = client.recv_message()
                self.assertEqual(r.envelope.routes[0].service, 'PreprocessorService:VanillaSlidingPreprocessor')
                for d in r.request.index.docs:
                    self.assertEqual(len(blob2array(d.chunks[0].blob).shape), 3)
                    self.assertEqual(blob2array(d.chunks[0].blob).shape[-1], 3)
                    self.assertEqual(blob2array(d.chunks[0].blob).shape[0], 224)
                    self.assertEqual(blob2array(d.chunks[0].blob).shape[1], 224)

    def test_segmentation_preprocessor_service_realdata(self):
        args = set_preprocessor_service_parser().parse_args([
            '--yaml_path', self.segmentation_img_pre_yaml
        ])

        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in)
        ])
        all_zips = zipfile.ZipFile(os.path.join(self.dirname, 'imgs/test.zip'))
        all_bytes = [all_zips.open(v).read() for v in all_zips.namelist()]

        with PreprocessorService(args), ZmqClient(c_args) as client:
            for req in RequestGenerator.index(all_bytes):
                msg = gnes_pb2.Message()
                msg.request.index.CopyFrom(req.index)
                client.send_message(msg)
                r = client.recv_message()
                self.assertEqual(r.envelope.routes[0].service, 'PreprocessorService:SegmentPreprocessor')
                for d in r.request.index.docs:
                    self.assertEqual(len(blob2array(d.chunks[0].blob).shape), 3)
                    self.assertEqual(blob2array(d.chunks[0].blob).shape[-1], 3)
                    self.assertEqual(blob2array(d.chunks[0].blob).shape[0], 224)
                    self.assertEqual(blob2array(d.chunks[0].blob).shape[1], 224)
                    print(blob2array(d.chunks[0].blob).dtype)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.




from typing import List

import numpy as np

from ..base import CompositionalTrainableBase, BaseTextEncoder
from ...helper import batching


class BertEncoder(BaseTextEncoder):
    store_args_kwargs = True
    is_trained = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._bc_encoder_args = args
        self._bc_encoder_kwargs = kwargs

    def post_init(self):
        from bert_serving.client import BertClient
        self.bc_encoder = BertClient(*self._bc_encoder_args, **self._bc_encoder_kwargs)

    @batching
    def encode(self, text: List[str], *args, **kwargs) -> np.ndarray:
        return self.bc_encoder.encode(text, *args, **kwargs)  # type: np.ndarray

    def close(self):
        self.bc_encoder.close()


class BertEncoderWithServer(CompositionalTrainableBase):
    def encode(self, text: List[str], *args, **kwargs) -> np.ndarray:
        return self.component['bert_client'].encode(text, *args, **kwargs)


class BertEncoderServer(BaseTextEncoder):
    store_args_kwargs = True
    is_trained = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        bert_args = ['-%s' % v for v in args]
        for k, v in kwargs.items():
            bert_args.append('-%s' % k)
            bert_args.append(str(v))
        self._bert_args = bert_args

    def post_init(self):
        from bert_serving.server import BertServer
        from bert_serving.server import get_args_parser
        self.bert_server = BertServer(get_args_parser().parse_args(self._bert_args))
        self.bert_server.start()
        self.bert_server.is_ready.wait()

    def close(self):
        self.bert_server.close()



import os
import unittest

from gnes.preprocessor.base import BasePreprocessor, PipelinePreprocessor
from gnes.proto import gnes_pb2


class P1(BasePreprocessor):
    def apply(self, doc: 'gnes_pb2.Document'):
        doc.doc_id += 1


class P2(BasePreprocessor):
    def apply(self, doc: 'gnes_pb2.Document'):
        doc.doc_id *= 3


class TestPartition(unittest.TestCase):
    def setUp(self):
        self.dirname = os.path.dirname(__file__)
        self.p3_name = 'pipe-p12'
        self.yml_dump_path = os.path.join(self.dirname, '%s.yml' % self.p3_name)
        self.bin_dump_path = os.path.join(self.dirname, '%s.bin' % self.p3_name)

    def tearDown(self):
        if os.path.exists(self.yml_dump_path):
            os.remove(self.yml_dump_path)
        if os.path.exists(self.bin_dump_path):
            os.remove(self.bin_dump_path)

    def test_pipelinepreproces(self):
        p3 = PipelinePreprocessor()
        p3.component = lambda: [P1(), P2()]
        d = gnes_pb2.Document()
        d.doc_id = 1
        p3.apply(d)
        self.assertEqual(d.doc_id, 6)

        p3.name = self.p3_name
        p3.dump_yaml()
        p3.dump()

        p4 = BasePreprocessor.load_yaml(p3.yaml_full_path)
        p4.apply(d)
        self.assertEqual(d.doc_id, 21)



import os
import unittest

from gnes.encoder.audio.mfcc import MfccEncoder
from gnes.preprocessor.helper import get_audio, get_video_length_from_raw


def extract_audio(video_bytes):
    sample_rate = 6500
    audio_interval = 1
    audios = []
    for raw_bytes in video_bytes:
        audio = get_audio(raw_bytes, sample_rate, audio_interval, get_video_length_from_raw(raw_bytes))
        audios.append(audio)
    audios = [slice for audio in audios for slice in audio]
    return audios


class TestMfccEncoder(unittest.TestCase):

    def setUp(self):
        self.dirname = os.path.dirname(__file__)
        self.video_path = os.path.join(self.dirname, 'videos')
        self.video_bytes = [open(os.path.join(self.video_path, _), 'rb').read()
                            for _ in os.listdir(self.video_path)]
        self.audios = extract_audio(self.video_bytes)
        self.mfcc_yaml = os.path.join(self.dirname, 'yaml', 'mfcc-encoder.yml')

    def test_mfcc_encoding(self):
        self.encoder = MfccEncoder.load_yaml(self.mfcc_yaml)
        vec = self.encoder.encode(self.audios)
        self.assertEqual(len(vec.shape), 2)
        self.assertEqual(vec.shape[0], len(self.audios))
        self.assertEqual(vec.shape[1] % self.encoder.n_mfcc, 0)


import os
import unittest

from gnes.encoder.text.gpt import GPTEncoder


# from numpy.testing import assert_array_equal


class TestGPTEncoder(unittest.TestCase):

    def setUp(self):
        dirname = os.path.dirname(__file__)
        self.dump_path = os.path.join(dirname, 'gpt_encoder.bin')
        self.test_str = []
        with open(os.path.join(dirname, 'sonnets.txt')) as f:
            for line in f:
                line = line.strip()
                if line:
                    self.test_str.append(line)

        self.gpt_encoder = GPTEncoder(
            model_dir=os.environ.get('GPT_CI_MODEL', '/openai_gpt'),
            pooling_strategy="REDUCE_MEAN")

    def test_encoding(self):

        vec = self.gpt_encoder.encode(self.test_str)
        self.assertEqual(vec.shape[0], len(self.test_str))
        self.assertEqual(vec.shape[1], 768)

    def test_dump_load(self):
        self.gpt_encoder.dump(self.dump_path)
        gpt_encoder2 = GPTEncoder.load(self.dump_path)

        vec = gpt_encoder2.encode(self.test_str)
        self.assertEqual(vec.shape[0], len(self.test_str))
        self.assertEqual(vec.shape[1], 768)

    def tearDown(self):
        if os.path.exists(self.dump_path):
            os.remove(self.dump_path)


import os
import unittest

from gnes.cli.parser import set_preprocessor_service_parser, _set_client_parser
from gnes.client.base import ZmqClient
from gnes.proto import gnes_pb2
from gnes.service.preprocessor import PreprocessorService


class TestProto(unittest.TestCase):

    def setUp(self):
        self.single_cn = '矫矫珍木巅，得无金丸惧。'
        self.single_en = 'When forty winters shall besiege thy brow. And dig deep trenches in thy beautys field.'
        self.dirname = os.path.dirname(__file__)
        self.yaml_path = os.path.join(self.dirname, 'yaml', 'test-preprocessor.yml')

    def test_preprocessor_service_empty(self):
        args = set_preprocessor_service_parser().parse_args(['--yaml_path', 'BasePreprocessor'])
        with PreprocessorService(args):
            pass

    def test_preprocessor_service_echo(self):
        args = set_preprocessor_service_parser().parse_args(['--yaml_path', 'BasePreprocessor'])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in)
        ])
        with PreprocessorService(args), ZmqClient(c_args) as client:
            msg = gnes_pb2.Message()
            msg.request.index.docs.extend([gnes_pb2.Document() for _ in range(5)])
            client.send_message(msg)
            r = client.recv_message()
            print(r)
            msg.request.train.docs.extend([gnes_pb2.Document() for _ in range(5)])
            client.send_message(msg)
            r = client.recv_message()
            print(r)

    def test_preprocessor_service_realdata(self):
        args = set_preprocessor_service_parser().parse_args(['--yaml_path', self.yaml_path])
        c_args = _set_client_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in)
        ])
        with open(os.path.join(self.dirname, '26-doc-chinese.txt'), 'r', encoding='utf8') as fp:
            msg = gnes_pb2.Message()
            all_text = ''
            for v in fp:
                if v.strip():
                    d = msg.request.train.docs.add()
                    d.raw_text = v
                    all_text += v
            with PreprocessorService(args), ZmqClient(c_args) as client:
                client.send_message(msg)
                r = client.recv_message()
                print(r)

                msg1 = gnes_pb2.Message()
                msg1.request.index.docs.extend(msg.request.train.docs)

                client.send_message(msg1)
                r = client.recv_message()
                print(r)

                msg2 = gnes_pb2.Message()
                msg2.request.search.query.raw_text = all_text

                client.send_message(msg2)
                r = client.recv_message()
                print(r)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from typing import List, Tuple, Any

import numpy as np

from ..base import BaseVectorIndexer
from ..key_only import ListKeyIndexer


class AnnoyIndexer(BaseVectorIndexer):

    def __init__(self, num_dim: int, data_path: str, metric: str = 'angular', n_trees=10, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_dim = num_dim
        self.data_path = data_path
        self.metric = metric
        self.n_trees = n_trees
        self._key_info_indexer = ListKeyIndexer()

    def post_init(self):
        from annoy import AnnoyIndex
        self._index = AnnoyIndex(self.num_dim, self.metric)
        try:
            if not os.path.exists(self.data_path):
                raise FileNotFoundError('"data_path" is not exist')
            if os.path.isdir(self.data_path):
                raise IsADirectoryError('"data_path" must be a file path, not a directory')
            self._index.load(self.data_path)
        except:
            self.logger.warning('fail to load model from %s, will create an empty one' % self.data_path)

    def add(self, keys: List[Tuple[int, Any]], vectors: np.ndarray, weights: List[float], *args, **kwargs):
        last_idx = self._key_info_indexer.size

        if len(vectors) != len(keys):
            raise ValueError('vectors length should be equal to doc_ids')

        if vectors.dtype != np.float32:
            raise ValueError("vectors should be ndarray of float32")

        for idx, vec in enumerate(vectors):
            self._index.add_item(last_idx + idx, vec)

        self._key_info_indexer.add(keys, weights)

    def query(self, keys: 'np.ndarray', top_k: int, *args, **kwargs) -> List[List[Tuple]]:
        self._index.build(self.n_trees)
        if keys.dtype != np.float32:
            raise ValueError('vectors should be ndarray of float32')
        res = []
        for k in keys:
            ret, relevance_score = self._index.get_nns_by_vector(k, top_k, include_distances=True)
            relevance_score = self.normalize_score(relevance_score, self.metric)
            chunk_info = self._key_info_indexer.query(ret)
            res.append([(*r, s) for r, s in zip(chunk_info, relevance_score)])
        return res

    def normalize_score(self, score: List[float], metrics: str, *args, **kwargs) -> List[float]:
        if metrics == 'angular':
            return list(map(lambda x:1 / (1 + x), score))
        elif metrics == 'euclidean':
            import math
            return list(map(lambda x:1 / (1 + math.sqrt(x) / self.num_dim), score))
        elif metrics == 'manhattan':
            return list(map(lambda x:1 / (1 + x / self.num_dim), score))
        elif metrics == 'hamming':
            return list(map(lambda x:1 / (1 + x), score))
        elif metrics == 'dot':
            raise NotImplementedError

    @property
    def size(self):
        return self._index.get_n_items()

    def __getstate__(self):
        d = super().__getstate__()
        self._index.save(self.data_path)
        return d



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.




import copy

import numpy as np

from ..base import BaseNumericEncoder
from ...helper import batching, train_required


class VladEncoder(BaseNumericEncoder):
    batch_size = 2048

    def __init__(self, num_clusters: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_clusters = num_clusters
        self.centroids = None

    def kmeans_train(self, vecs):
        import faiss

        kmeans = faiss.Kmeans(vecs.shape[1], self.num_clusters, niter=5, verbose=False)
        kmeans.train(vecs)
        self.centroids = kmeans.centroids

    def kmeans_pred(self, vecs):
        vecs = np.reshape(vecs, [vecs.shape[0], 1, 1, vecs.shape[1]])
        dist = np.sum(np.square(vecs - self.centroids), -1)
        return np.argmax(-dist, axis=-1).astype(np.int64)

    @batching
    def train(self, vecs: np.ndarray, *args, **kwargs):
        assert len(vecs) > self.num_clusters, 'number of data should be larger than number of clusters'
        vecs_ = copy.deepcopy(vecs)
        vecs_ = np.concatenate((list(vecs_[i] for i in range(len(vecs_)))), axis=0)
        self.kmeans_train(vecs_)

    @train_required
    @batching
    def encode(self, vecs: np.ndarray, *args, **kwargs) -> np.ndarray:
        vecs_ = copy.deepcopy(vecs)
        vecs_ = np.concatenate((list(vecs_[i] for i in range(len(vecs_)))), axis=0)

        knn_output = self.kmeans_pred(vecs_)
        knn_output = [knn_output[i:i + vecs.shape[1]] for i in range(0, len(knn_output), vecs.shape[1])]

        output = []
        for chunk_count, chunk in enumerate(vecs):
            res = np.zeros((self.centroids.shape[0], self.centroids.shape[1]))
            for frame_count, frame in enumerate(chunk):
                center_index = knn_output[chunk_count][frame_count][0]
                res[center_index] += (frame - self.centroids[center_index])
            output.append(res)

        output = np.array(list(map(lambda x: x.reshape(1, -1), output)), dtype=np.float32)
        output = np.squeeze(output, axis=1)
        return output

    def _copy_from(self, x: 'VladEncoder') -> None:
        self.num_clusters = x.num_clusters
        self.centroids = x.centroids



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import io
from typing import List

import numpy as np
from PIL import Image

from .base import BaseImagePreprocessor
from ..video.ffmpeg import FFmpegPreprocessor
from ...proto import gnes_pb2, array2blob


class BaseSlidingPreprocessor(BaseImagePreprocessor):

    def __init__(self, window_size: int = 64,
                 stride_height: int = 64,
                 stride_wide: int = 64,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.window_size = window_size
        self.stride_height = stride_height
        self.stride_wide = stride_wide

    def apply(self, doc: 'gnes_pb2.Document'):
        super().apply(doc)
        if doc.raw_bytes:
            original_image = Image.open(io.BytesIO(doc.raw_bytes))
            all_subareas, index = self._get_all_subarea(original_image)
            image_set, center_point_list = self._get_all_sliding_window(np.array(original_image))
            normalizaed_image_set = [np.array(self._torch_transform(img)).transpose(1, 2, 0)
                                     for img in image_set]
            weight = self._get_all_chunks_weight(normalizaed_image_set)

            for ci, ele in enumerate(zip(normalizaed_image_set, weight)):
                c = doc.chunks.add()
                c.doc_id = doc.doc_id
                c.blob.CopyFrom(array2blob(ele[0]))
                c.offset_1d = ci
                c.offset_nd.x.extend(self._get_slid_offset_nd(all_subareas, index, center_point_list[ci]))
                c.weight = ele[1]
        else:
            self.logger.error('bad document: "raw_bytes" is empty!')

    def _get_all_sliding_window(self, img: 'np.ndarray'):
        extend_height = self.window_size - (img.shape[0]) % self.stride_height
        extend_wide = self.window_size - (img.shape[1]) % self.stride_wide

        input = np.pad(img, ((0, extend_height),
                             (0, extend_wide),
                             (0, 0)),
                       mode='constant', constant_values=0)
        expanded_input = np.lib.stride_tricks.as_strided(
            input,
            shape=(
                1 + int((input.shape[0] - self.window_size) / self.stride_height),
                1 + int((input.shape[1] - self.window_size) / self.stride_wide),
                self.window_size,
                self.window_size,
                3
            ),
            strides=(
                input.strides[0] * self.stride_height,
                input.strides[1] * self.stride_wide,
                input.strides[0],
                input.strides[1],
                1
            ),
            writeable=False
        )
        center_point_list = [
            [self.window_size / 2 + x * self.stride_wide, self.window_size / 2 + y * self.stride_height]
            for x in range(expanded_input.shape[0])
            for y in range(expanded_input.shape[1])]

        expanded_input = expanded_input.reshape((-1, self.window_size, self.window_size, 3))
        return [np.array(Image.fromarray(img).resize((self.target_img_size, self.target_img_size))) for img in
                expanded_input], center_point_list

    def _get_slid_offset_nd(self, all_subareas: List[List[int]], index: List[List[int]], center_point: List[float]) -> \
            List[int]:
        location_list = self._get_location(all_subareas, center_point)
        location = [i for i in range(len(location_list)) if location_list[i] is True][0]
        return index[location][:2]

    @staticmethod
    def _get_location(all_subareas: List[List[int]], center_point: List[float]) -> List[bool]:
        location_list = []
        x_boundary = max([x[2] for x in all_subareas])
        y_boundary = max([y[3] for y in all_subareas])
        for area in all_subareas:
            if center_point[0] in range(int(area[0]), int(area[2])) and center_point[1] in range(int(area[1]),
                                                                                                 int(area[3])):
                location_list.append(True)
            elif center_point[0] in range(int(area[0]), int(area[2])) and y_boundary == area[3] and center_point[
                1] > y_boundary:
                location_list.append(True)
            elif center_point[1] in range(int(area[1]), int(area[3])) and x_boundary == area[2] and center_point[
                0] > x_boundary:
                location_list.append(True)
            else:
                location_list.append(False)
        if True not in location_list:
            location_list[-1] = True
        return location_list


class VanillaSlidingPreprocessor(BaseSlidingPreprocessor):

    def _get_all_chunks_weight(self, image_set: List['np.ndarray']) -> List[float]:
        return [1 / len(image_set) for _ in range(len(image_set))]


class WeightedSlidingPreprocessor(BaseSlidingPreprocessor):

    def _get_all_chunks_weight(self, image_set: List['np.ndarray']) -> List[float]:
        return FFmpegPreprocessor.pic_weight(image_set)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from typing import List, Callable

import numpy as np

from ..base import BaseImageEncoder
from ...helper import batching


class BasePytorchEncoder(BaseImageEncoder):
    batch_size = 64

    def __init__(self, model_name: str,
                 layers: List[str],
                 model_dir: str,
                 use_cuda: bool = False,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.model_dir = model_dir
        self.model_name = model_name
        self.layers = layers
        self._use_cuda = use_cuda

    def post_init(self):
        import torch
        import torchvision.models as models

        class _Model(torch.nn.Module):
            def __init__(self, model_name: str, layers: List[str]):
                super().__init__()

                self.m = getattr(models, model_name)(pretrained=True)
                self.layers = [self.fn_parser(l) for l in layers]

            def fn_parser(self, layer: str) -> Callable:

                if '(' not in layer and ')' not in layer:
                    # this is a shorthand syntax we need to add "(x)" at the end
                    layer = 'm.%s(x)' % layer
                else:
                    pass

                def layer_fn(x, l, m, torch):
                    return eval(l)

                return lambda x: layer_fn(x, layer, self.m, torch)

            def forward(self, x):
                for l in self.layers:
                    x = l(x)
                return x

        os.environ['TORCH_HOME'] = self.model_dir
        self._model = _Model(self.model_name, self.layers)
        self._model = self._model.eval()
        if self._use_cuda:
            # self._model.cuda()
            self._device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
            self._model = self._model.to(self._device)

    @batching
    def encode(self, img: List['np.ndarray'], *args, **kwargs) -> np.ndarray:
        import torch
        self._model.eval()

        # padding to ensure that every chunk has the same number of frame
        def _padding(img: List['np.ndarray']):
            max_lenth = max([len(x) for x in img])
            img = [np.concatenate((im, np.zeros((max_lenth - im.shape[0], im.shape[1], im.shape[2], 3), dtype=np.uint8))
                                  , axis=0)
                   if im.shape[0] < max_lenth else im for im in img]
            return img, max_lenth

        @batching
        def _encode(_, img: List['np.ndarray']):
            import copy

            if len(img[0].shape) == 4:
                img_ = copy.deepcopy(img)
                img_ = np.concatenate((list(img_[i] for i in range(len(img_)))), axis=0)
                img_for_torch = np.array(img_, dtype=np.float32).transpose(0, 3, 1, 2)
            else:
                img_for_torch = np.array(img, dtype=np.float32).transpose(0, 3, 1, 2)

            img_tensor = torch.from_numpy(img_for_torch)
            if self._use_cuda:
                img_tensor = img_tensor.cuda()

            result_npy = []
            for t in img_tensor:
                t = torch.unsqueeze(t, 0)
                encodes = self._model(t)
                encodes = torch.squeeze(encodes, 0)
                result_npy.append(encodes.data.cpu().numpy())

            output = np.array(result_npy, dtype=np.float32)

            if len(img[0].shape) == 4:
                output = output.reshape((len(img), max_lenth, -1))
            return output

        # for video
        if len(img[0].shape) == 4:
            padding_image, max_lenth = _padding(img)
            output = _encode(self, padding_image)
        # for image
        else:
            output = _encode(self, img)

        return output



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


# A key-value map for Class to the (module)file it located in
from ..base import register_all_class

_cls2file_map = {
    'BaseRouter': 'base',
    'BaseMapRouter': 'base',
    'BaseReduceRouter': 'base',
    'ChunkToDocumentRouter': 'reduce.chunk',
    'DocFillRouter': 'reduce.document',
    'ConcatEmbedRouter': 'reduce.concat',
    'PublishRouter': 'map.simple',
    'DocBatchRouter': 'map.simple',
    'ChunkSumRouter': 'reduce.chunk_sum',
    'DocSumRouter': 'reduce.doc_sum'

}

register_all_class(_cls2file_map, 'router')



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re

from .base import BaseTextPreprocessor
from ...proto import gnes_pb2


class TextPreprocessor(BaseTextPreprocessor):
    def __init__(self, deliminator: str = r'[.。！？!?]+', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.deliminator = deliminator

    def apply(self, doc: 'gnes_pb2.Document') -> None:
        super().apply(doc)
        doc.raw_text = doc.raw_bytes.decode().strip()
        for ci, s in enumerate(re.split(self.deliminator, doc.raw_text)):
            if s.strip():
                c = doc.chunks.add()
                c.doc_id = doc.doc_id
                c.text = s.strip()
                c.offset_1d = ci
                c.weight = len(c.text) / len(doc.raw_text)



import os
import unittest

from gnes.indexer.base import BaseIndexer


class TestMHIndexer(unittest.TestCase):

    def setUp(self):
        dirname = os.path.dirname(__file__)
        self.yaml_path1 = os.path.join(dirname, 'yaml', 'base-indexer2.yml')
        self.yaml_path2 = os.path.join(dirname, 'yaml', 'base-indexer3.yml')
        self.yaml_path3 = os.path.join(dirname, 'yaml', 'base-indexer4.yml')

    def test_ind1(self):
        self.assertRaises(ValueError, BaseIndexer.load_yaml, self.yaml_path1)

    def test_ind3(self):
        self.assertRaises(ValueError, BaseIndexer.load_yaml, self.yaml_path1)

    def test_ind2(self):
        mhi = BaseIndexer.load_yaml(self.yaml_path2)
        print(mhi)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.




import pickle
from threading import Thread, Event
from typing import List, Any

from ..base import BaseTextIndexer
from ...proto import gnes_pb2


class LVDBIndexer(BaseTextIndexer):

    def __init__(self, data_path: str,
                 keep_na_doc: bool = True,
                 drop_raw_bytes: bool = False,
                 drop_chunk_blob: bool = False,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_path = data_path
        self.keep_na_doc = keep_na_doc
        self.drop_raw_bytes = drop_raw_bytes
        self.drop_chunk_blob = drop_chunk_blob
        self._NOT_FOUND = None

    def post_init(self):
        import plyvel
        self._db = plyvel.DB(self.data_path, create_if_missing=True)

    def add(self, keys: List[int], docs: List['gnes_pb2.Document'], *args, **kwargs):
        with self._db.write_batch() as wb:
            for k, d in zip(keys, docs):
                doc_id = pickle.dumps(k)
                if self.drop_raw_bytes:
                    d.raw_bytes = b''
                if self.drop_chunk_blob:
                    for i in range(len(d.chunks)):
                        d.chunks[i].ClearField('blob')
                doc = d.SerializeToString()
                wb.put(doc_id, doc)

    def query(self, keys: List[int], *args, **kwargs) -> List['gnes_pb2.Document']:
        res = []
        for k in keys:
            doc_id = pickle.dumps(k)
            v = self._db.get(doc_id)
            doc = gnes_pb2.Document()
            if v is not None:
                doc.ParseFromString(v)
                res.append(doc)
            elif self.keep_na_doc:
                res.append(self._NOT_FOUND)
        return res

    def close(self):
        super().close()
        self._db.close()


class AsyncLVDBIndexer(LVDBIndexer):
    def post_init(self):
        super().post_init()
        self._is_listening = Event()
        self._is_listening.set()
        self._is_idle = Event()
        self._is_idle.set()
        self._jobs = []
        self._thread = Thread(target=self._db_write)
        self._thread.setDaemon(1)
        self._thread.start()

    def add(self, keys: List[int], docs: List['gnes_pb2.Document'], *args, **kwargs):
        self._jobs.append((keys, docs))

    def query(self, *args, **kwargs) -> List[Any]:
        self._is_idle.wait()
        return super().query(*args, **kwargs)

    def _db_write(self):
        while self._is_listening.is_set():
            while self._jobs:
                self._is_idle.clear()
                keys, docs = self._jobs.pop()
                super().add(keys, docs)
            self._is_idle.set()

    def close(self):
        self._jobs.clear()
        self._is_listening.clear()
        self._is_idle.wait()
        self._thread.join()
        super().close()



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from typing import List, Union

from .base import BaseService as BS, MessageHandler
from ..proto import gnes_pb2, array2blob, blob2array


class EncoderService(BS):
    handler = MessageHandler(BS.handler)

    def post_init(self):
        from ..encoder.base import BaseEncoder
        self._model = self.load_model(BaseEncoder)
        self.train_data = []

    @staticmethod
    def get_chunks_from_docs(docs: Union[List['gnes_pb2.Document'], 'gnes_pb2.Document']) -> List:
        if getattr(docs, 'doc_type', None) is not None:
            docs = [docs]
        return [c.text if d.doc_type == gnes_pb2.Document.TEXT else blob2array(c.blob)
                for d in docs for c in d.chunks]

    @handler.register(gnes_pb2.Request.IndexRequest)
    def _handler_index(self, msg: 'gnes_pb2.Message'):
        chunks = self.get_chunks_from_docs(msg.request.index.docs)
        vecs = self._model.encode(chunks)
        s = 0
        for d in msg.request.index.docs:
            if len(d.chunks):
                d.chunk_embeddings.CopyFrom(array2blob(vecs[s:(s + len(d.chunks))]))
                s += len(d.chunks)

    @handler.register(gnes_pb2.Request.TrainRequest)
    def _handler_train(self, msg: 'gnes_pb2.Message'):
        if msg.request.train.docs:
            chunks = self.get_chunks_from_docs(msg.request.train.docs)
            self.train_data.extend(chunks)
            msg.response.train.status = gnes_pb2.Response.PENDING
            # raise BlockMessage
        if msg.request.train.flush:
            self._model.train(self.train_data)
            self.logger.info('%d samples is flushed for training' % len(self.train_data))
            self.is_model_changed.set()
            self.train_data.clear()
            msg.response.control.status = gnes_pb2.Response.SUCCESS

    @handler.register(gnes_pb2.Request.QueryRequest)
    def _handler_search(self, msg: 'gnes_pb2.Message'):
        chunks = self.get_chunks_from_docs(msg.request.search.query)
        vecs = self._model.encode(chunks)
        msg.request.search.query.chunk_embeddings.CopyFrom(array2blob(vecs))



import os
import unittest

import numpy as np

from gnes.indexer.vector.bindexer import BIndexer


class TestBIndexer(unittest.TestCase):
    def setUp(self):
        self.toy_data = np.array([[1, 2, 1, 2],
                                  [2, 1, 3, 4],
                                  [1, 2, 1, 2],
                                  [2, 1, 4, 3],
                                  [2, 1, 3, 4],
                                  [23, 32, 21, 33],
                                  [123, 132, 1, 1]]).astype(np.uint8)
        self.toy_label = [(234, 0), (432, 0), (123, 1), (321, 0), (1, 0), (2, 0), (6, 0)]

        self.toy_query = np.array([[1, 2, 1, 2],
                                   [2, 1, 3, 4],
                                   [3, 2, 1, 2]]).astype(np.uint8)

        self.toy_exp = [[(234, 0, 1., 1,), (123, 1, 1., 1)], [(432, 0, 1., 1), (1, 0, 1., 1)],
                         [(234, 0, 1., 0.75), (123, 1, 1., 0.75)]]
        self.weights = [1.] * len(self.toy_label)

        dirname = os.path.dirname(__file__)
        self.dump_path = os.path.join(dirname, 'test-indexer.bin')

    def tearDown(self):
        if os.path.exists(self.dump_path):
            os.remove(self.dump_path)

    def test_nsw_search(self):
        fd = BIndexer(self.toy_data.shape[1], data_path=self.dump_path + '_1')
        fd.add(self.toy_label, self.toy_data, self.weights)

        rs = fd.query(self.toy_query, 2, method='nsw', normalized_score=False)
        for i in range(len(rs)):
            rs[i] = sorted(rs[i], key=lambda x: (x[3], -x[0]))
        fd.close()
        self.assertEqual(rs, self.toy_exp)

    def test_force_search(self):
        fd = BIndexer(self.toy_data.shape[1], data_path=self.dump_path + '_2')
        fd.add(self.toy_label, self.toy_data, self.weights)
        rs = fd.query(self.toy_query, 2, method='force', normalized_score=False)
        for i in range(len(rs)):
            rs[i] = sorted(rs[i], key=lambda x: (x[3], -x[0]))
        fd.close()
        self.assertEqual(rs, self.toy_exp)

    def test_dump_load(self):
        fd = BIndexer(self.toy_data.shape[1], data_path=self.dump_path + '_3')
        fd.add(self.toy_label, self.toy_data, self.weights)
        fd.dump()
        fd.close()
        # shutil.rmtree(self.data_path + "_3")

        fd2 = BIndexer.load(fd.dump_full_path)
        rs = fd2.query(self.toy_query, 2, normalized_score=False)
        for i in range(len(rs)):
            rs[i] = sorted(rs[i], key=lambda x: (x[3], -x[0]))
        fd2.close()

        self.assertEqual(rs, self.toy_exp)
        os.remove(self.dump_path + '_3')



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List

import numpy as np
from PIL import Image

from ..base import BaseImageEncoder


class CVAEEncoder(BaseImageEncoder):
    batch_size = 64

    def __init__(self, model_dir: str,
                 latent_dim: int = 300,
                 select_method: str = 'MEAN',
                 l2_normalize: bool = False,
                 use_gpu: bool = True,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.model_dir = model_dir
        self.latent_dim = latent_dim
        self.select_method = select_method
        self.l2_normalize = l2_normalize
        self.use_gpu = use_gpu

    def post_init(self):
        import tensorflow as tf
        from .cvae_cores.model import CVAE
        g = tf.Graph()
        with g.as_default():
            self._model = CVAE(self.latent_dim)
            self.inputs = tf.placeholder(tf.float32,
                                         (None, 120, 120, 3))

            self.mean, self.var = self._model.encode(self.inputs)

            config = tf.ConfigProto(log_device_placement=False)
            if self.use_gpu:
                config.gpu_options.allow_growth = True
            self.sess = tf.Session(config=config)
            self.saver = tf.train.Saver()
            self.saver.restore(self.sess, self.model_dir)

    def encode(self, img: List['np.ndarray'], *args, **kwargs) -> np.ndarray:
        img = [(np.array(Image.fromarray(im).resize((120, 120)),
                         dtype=np.float32) / 255) for im in img]

        def _encode(_, data):
            _mean, _var = self.sess.run((self.mean, self.var),
                                        feed_dict={self.inputs: data})
            if self.select_method == 'MEAN':
                return _mean
            elif self.select_method == 'VAR':
                return _var
            elif self.select_method == 'MEAN_VAR':
                return np.concatenate([_mean, _var], axis=1)
            else:
                raise NotImplementedError

        v = _encode(None, img).astype(np.float32)
        if self.l2_normalize:
            v = v / (v ** 2).sum(axis=1, keepdims=True) ** 0.5
        return v



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import numpy as np
import tensorflow as tf


class CVAE(tf.keras.Model):
    def __init__(self, latent_dim):
        super(CVAE, self).__init__()
        self.latent_dim = latent_dim
        self.inference_net = tf.keras.Sequential(
            [
                tf.keras.layers.InputLayer(input_shape=(120, 120, 3)),
                tf.keras.layers.Conv2D(
                    filters=32, kernel_size=3, strides=(2, 2),
                    padding='SAME',
                    activation='relu'),
                tf.keras.layers.Conv2D(
                    filters=32, kernel_size=3, strides=(2, 2),
                    padding='SAME',
                    activation='relu'),
                tf.keras.layers.Conv2D(
                    filters=32, kernel_size=3, strides=(2, 2),
                    padding='SAME',
                    activation='relu'),
                tf.keras.layers.Flatten(),
                # No activation
                tf.keras.layers.Dense(latent_dim + latent_dim),
            ]
        )

        self.generative_net = tf.keras.Sequential(
            [
                tf.keras.layers.InputLayer(input_shape=(latent_dim,)),
                tf.keras.layers.Dense(units=15 * 15 * 32,
                                      activation=tf.nn.relu),
                tf.keras.layers.Reshape(target_shape=(15, 15, 32)),
                tf.keras.layers.Conv2DTranspose(
                    filters=32,
                    kernel_size=3,
                    strides=(2, 2),
                    padding="SAME",
                    activation='relu'),
                tf.keras.layers.Conv2DTranspose(
                    filters=32,
                    kernel_size=3,
                    strides=(2, 2),
                    padding="SAME",
                    activation='relu'),
                tf.keras.layers.Conv2DTranspose(
                    filters=32,
                    kernel_size=3,
                    strides=(2, 2),
                    padding="SAME",
                    activation='relu'),
                # No activation
                tf.keras.layers.Conv2DTranspose(
                    filters=3, kernel_size=3, strides=(1, 1), padding="SAME"),
            ]
        )

    def sample(self, eps=None):
        if eps is None:
            eps = tf.random_normal(shape=(100, self.latent_dim))
        return self.decode(eps, apply_sigmoid=True)

    def encode(self, x):
        mean, logvar = tf.split(self.inference_net(x), num_or_size_splits=2, axis=1)
        return mean, logvar

    @staticmethod
    def reparameterize(mean, logvar):
        eps = tf.random_normal(shape=tf.shape(mean))
        return eps * tf.exp(logvar * .5) + mean

    def decode(self, z, apply_sigmoid=False):
        logits = self.generative_net(z)
        if apply_sigmoid:
            probs = tf.sigmoid(logits)
            return probs

        return logits

    def compute_loss(self, x):
        mean, logvar = self.encode(x)
        z = CVAE.reparameterize(mean, logvar)
        x_logit = self.decode(z)

        cross_ent = tf.nn.sigmoid_cross_entropy_with_logits(logits=x_logit,
                                                            labels=x)
        logpx_z = -tf.reduce_sum(cross_ent, axis=[1, 2, 3])
        logpz = CVAE.log_normal_pdf(z, 0., 0.)
        logqz_x = CVAE.log_normal_pdf(z, mean, logvar)

        return -tf.reduce_mean(logpx_z + logpz - logqz_x)

    @staticmethod
    def log_normal_pdf(sample, mean, logvar, raxis=1):
        log2pi = tf.math.log(2. * np.pi)
        return tf.reduce_sum(
            -.5 * ((sample - mean) ** 2. * tf.exp(-logvar) + logvar + log2pi),
            axis=raxis)



import os
import shutil
import unittest

import numpy as np

from gnes.helper import touch_dir
from gnes.indexer.vector.annoy import AnnoyIndexer


class TestAnnoyIndexer(unittest.TestCase):
    def setUp(self):
        self.toy_data = np.random.random([10, 5]).astype(np.float32)

        dirname = os.path.dirname(__file__)
        self.dump_path = os.path.join(dirname, 'indexer.pkl')

    def tearDown(self):
        if os.path.exists(self.dump_path):
            os.remove(self.dump_path)

    def test_search(self):
        a = AnnoyIndexer(5, self.dump_path)
        a.add(list(zip(list(range(10)), list(range(10)))), self.toy_data, [1.] * 10)
        self.assertEqual(a.size, 10)
        top_1 = [i[0][0] for i in a.query(self.toy_data, top_k=1)]
        self.assertEqual(top_1, list(range(10)))
        a.close()



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List

from ..base import BasePreprocessor
from ...proto import gnes_pb2


class BaseImagePreprocessor(BasePreprocessor):
    doc_type = gnes_pb2.Document.IMAGE

    def __init__(self,
                 target_img_size: int = 224,
                 is_rgb: bool = True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.target_img_size = target_img_size
        self.is_rgb = is_rgb

    def _get_all_chunks_weight(self, image_set: List['np.ndarray']) -> List[float]:
        pass

    @staticmethod
    def _torch_transform(image):
        import torchvision.transforms as transforms
        return transforms.Compose([transforms.ToTensor(),
                                   transforms.Normalize(mean=(0.5, 0.5, 0.5), std=(0.5, 0.5, 0.5))])(image)

    @staticmethod
    def _get_all_subarea(image):
        from itertools import product
        x_list = [0, image.size[0] / 3, 2 * image.size[0] / 3, image.size[0]]
        y_list = [0, image.size[1] / 3, 2 * image.size[1] / 3, image.size[1]]

        index = [[x, y, x + 1, y + 1] for [x, y] in product(range(len(x_list) - 1), range(len(y_list) - 1))]
        all_subareas = [[x_list[idx[0]], y_list[idx[1]], x_list[idx[2]], y_list[idx[3]]] for idx in index]
        return all_subareas, index



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


def _start_service(cls, args):
    from ..service.base import ServiceManager
    with ServiceManager(cls, args) as es:
        es.join()


def grpc(args):
    from ..service.grpc import GRPCService
    _start_service(GRPCService, args)


def preprocess(args):
    from ..service.preprocessor import PreprocessorService
    _start_service(PreprocessorService, args)


def encode(args):
    from ..service.encoder import EncoderService
    _start_service(EncoderService, args)


def index(args):
    from ..service.indexer import IndexerService
    _start_service(IndexerService, args)


def route(args):
    from ..service.router import RouterService
    _start_service(RouterService, args)


def frontend(args):
    from gnes.service.frontend import FrontendService
    import threading
    with FrontendService(args):
        forever = threading.Event()
        forever.wait()


def client_http(args):
    from ..client.http import HttpClient
    HttpClient(args).start()


def client_cli(args):
    from ..client.cli import CLIClient
    CLIClient(args)


def compose(args):
    from ..composer.base import YamlComposer
    from ..composer.flask import YamlComposerFlask

    if args.flask:
        YamlComposerFlask(args).run()
    else:
        YamlComposer(args).build_all()



import os
import unittest

import numpy as np

from gnes.helper import batch_iterator, TimeContext
from gnes.indexer.key_only import DictKeyIndexer, NumpyKeyIndexer, ListKeyIndexer, ListNumpyKeyIndexer


class TestProto(unittest.TestCase):
    def setUp(self):
        self.num_sample = 1000000
        self.num_query = 10000
        self.keys = np.array([j for j in range(self.num_sample)])
        self.key_offset = np.stack([self.keys, np.random.randint(0, 255, size=[self.num_sample])],
                                   axis=1).tolist()
        self.weights = np.random.random(size=[self.num_sample]).tolist()
        self.query = np.random.randint(0, self.num_sample, size=[self.num_query]).tolist()
        self.dump_path = './dump.bin'

    def tearDown(self):
        if os.path.exists(self.dump_path):
            os.remove(self.dump_path)

    def _test_any(self, cls):
        a = cls()
        self.assertEqual(a.size, 0)
        a.add(self.key_offset, self.weights)
        self.assertEqual(a.size, self.num_sample)

        res1 = a.query(self.query)
        res2 = [(*self.key_offset[q], self.weights[q]) for q in self.query]
        self.assertListEqual(res1, res2)

        # test dump and reload
        a.dump(self.dump_path)
        b = cls.load(self.dump_path)
        res1 = a.query(self.query)
        res2 = b.query(self.query)
        self.assertListEqual(res1, res2)

    def test_numpy(self):
        self._test_any(NumpyKeyIndexer)

    def test_list(self):
        self._test_any(ListKeyIndexer)

    def test_fast_list(self):
        self._test_any(ListNumpyKeyIndexer)

    def test_dict(self):
        self._test_any(DictKeyIndexer)

    def test_bench_numpy_list(self):
        for cls in [ListKeyIndexer, NumpyKeyIndexer, ListNumpyKeyIndexer, DictKeyIndexer]:
            a = cls()
            b_size = 1000
            with TimeContext('%s:add()' % cls.__name__):
                for k, w in zip(batch_iterator(self.key_offset, b_size), batch_iterator(self.weights, b_size)):
                    a.add(k, w)

            with TimeContext('%s:query()' % cls.__name__):
                for k in batch_iterator(self.query, b_size):
                    a.query(k)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from .base import BaseService as BS, MessageHandler
from ..proto import gnes_pb2


class PreprocessorService(BS):
    handler = MessageHandler(BS.handler)

    def post_init(self):
        from ..preprocessor.base import BasePreprocessor
        self._model = self.load_model(BasePreprocessor)

    @handler.register(gnes_pb2.Request.TrainRequest)
    def _handler_train(self, msg: 'gnes_pb2.Message'):
        for d in msg.request.train.docs:
            self._model.apply(d)

    @handler.register(gnes_pb2.Request.IndexRequest)
    def _handler_index(self, msg: 'gnes_pb2.Message'):
        for d in msg.request.index.docs:
            self._model.apply(d)

    @handler.register(gnes_pb2.Request.QueryRequest)
    def _handler_query(self, msg: 'gnes_pb2.Message'):
        self._model.apply(msg.request.search.query)
